# solves the equation l[1]*y(z)+l[2]*diff(y(z),z)+...+l[n+1]*diff(y(z),z$n)+l[n+2]=0 #

ode::solve_linear := proc(l,z,n)
local inho,i,k,sys,up,y,partsol,lc,gensol,yy,Y;
# global ode::depth; #
begin
   if ode::depth>ode::maxDepth then return(FAIL) end_if;
   userinfo(1,"linear ordinary differential equation of order ".n);
   if n=0 then # l[1]*y(z)+l[2]=0 #
      if iszero(l[1]) then return({}) else return({-l[2]/l[1]}) end_if
   end_if;
   userinfo(2,_plus(l[i+1]*diff(Y(z),z$i)$i=0..n)+l[n+2]=0);
   # try first exact_nth_order which deals with the inhomogeneous coeff. #
   y:=ode::exact_nth_order(l,z,n);
   if y<>FAIL then return(y) end_if;
   inho:=l[n+2]; l[n+2]:=NIL; lc:=l[n+1]; # leading coefficient #
   y:=FAIL;
   if not has(l,z) then
      userinfo(1,"with constant coefficients");
      y:=ode::constant(l,z,n)
   else
      if n=1 then y:=[ode::firstord(l[1]*yy(z)+l[2]*diff(yy(z),z),yy,z)]
      else
         if map({op(l)},testtype,Type::PolyExpr(z,Type::Rational))={TRUE} then
            userinfo(1,"with polynomial coefficients");
            # removed temporarily because of bug with
              diff(y(x), x $ 2)-2/x*diff(y(x), x) + 3/x^2*y(x)-2*x+1 #
            # y:=ode::expsols(l,Y,z,n); #
            if nops(y)<n then y:=ode::euler(l,z,n) end_if;
         else
            userinfo(1,"with non constant coefficients");
         end_if;
      end_if;
   end_if;
   userinfo(2,"solution of homogeneous part is",y);
   if y=FAIL then return(ode::adjoint(l,z,inho)) end_if;
   gensol:=_plus(genident("C")*op(y,i)$hold(i)=1..n);
   # a solution of the homogeneous part of the linear DE is C1*y[1]+...+Cn*y[n] #
   # we now find a particular solution of the full equation #
   # using the method described on page 314 of Zwillinger (Variation of parameters) #
   if inho=0 then partsol:=0
   else
      sys:={_plus(diff(y[k],z$i)*up[k] $ k=1..n) $ hold(i)=0..n-2,
		_plus(diff(y[k],z$(n-1))*up[k] $ k=1..n)+inho/lc},{up[k] $ k=1..n};
      userinfo(2,"system to solve is",sys);
      sys:=linsolve(sys);
      if sys=FAIL then return(FAIL) end_if;
      userinfo(2,"solution is",sys);
      for k from 1 to n do
         userinfo(2,"integrating",up[k]);
         yy[k]:=int(subs(up[k],sys),z);
         userinfo(2,"solution is",yy[k]);
      end_for;
      partsol:=_plus(yy[k]*y[k] $ hold(k)=1..n);
   end_if;
   partsol:=expand(partsol);
   if has(partsol,{sin,cos}) then
      k:=combine(partsol,sincos);
      if k<>partsol and has(k,hold(int)) then k:=level(k) end_if;
      partsol:=k
   end_if;
   {gensol+partsol}
end_proc:

ode::constant := proc(l,z,n) local char_eq,i,X;
begin
   char_eq:=_plus(l[i+1]*X^i $ i=0..n);
   userinfo(2,"solving characteristic equation",char_eq);
   l:=solve(char_eq,X,hold(Multiple));
   l:=[expand(l)];
   userinfo(2,"solutions are",l);
   if nops(l)<>n then return(FAIL) end_if;
   ode::vspace(l,z);
end_proc:
